#include <avr/io.h>

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

#include <util/delay.h>

// video timing
#define LINE_TIME 1271
#define SLEEP_TIME 1250

// screen layout
#define V_TOP		55
#define V_HEADER	55
#define V_HEADER_SP 64
#define V_HEADER_R	65
#define V_GRAPH		66
#define V_FOOTER_R	194
#define V_FOOTER_SP 195
#define V_FOOTER	196
#define V_BOTTOM	205

// screen sections
#define S_BLANK		0
#define S_TEXT		1
#define S_GRAPH		2
#define S_RULE		3

// video generation
uint8_t sync_on, sync_off;
uint16_t video_line;
uint8_t char_line;
uint16_t graph_offset;
uint8_t section;
uint8_t* text;

// cursor and trigger
uint8_t x_cursor;
uint8_t y_trigger;

// header and footer
uint8_t header[26] = "                         \0";
uint8_t footer[26] = "                         \0";

// bitmask for drawing points
uint8_t bitmask[8] PROGMEM = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};

// graph raster
uint8_t graph[3200];

// array of samples
uint8_t samples[200];

/*
 * TIMER2_COMPB: Puts the microcontroller to sleep before new scanline
 */
ISR(TIMER1_COMPB_vect, ISR_NAKED)
{
	sei();
	sleep_cpu();
	reti();
}

/*
 * TIMER1_COMPA: Generates sync pulses and scanlines.  Must be entered from
 * 		sleep mode for consistent timing
 */
ISR(TIMER1_COMPA_vect)
{
	uint8_t out, x;

	// start the horizontal sync pulse
	PORTD = (PORTD & ~(_BV(PORTD2))) | sync_on;

	// begin inverted (vertical) sync after line 247
	if (video_line==247)
	{
		sync_on = _BV(PORTD2);
		sync_off = 0;
	}

	// back to regular sync after line 250
	if (video_line==250)
	{
		sync_on = 0;
		sync_off = _BV(PORTD2);
	}

	// adjust to make 5 us pulses
	_delay_us(4);

	// end sync pulse
	PORTD = (PORTD & ~(_BV(PORTD2))) | sync_off;

	// don't send data to the screen for non-visible portions of the screen,
	// this allows more time for the main loop to execute
	if (video_line < V_BOTTOM && video_line >= V_TOP)
	{
		// adjust to center the image
		_delay_us(7);

		// send an initial blank byte
		if (section != S_BLANK)
			UDR0 = 0xFF;

		// send data data depending on screen section
		switch (section)
		{
		case S_TEXT:
			// draw character data
			for (x = 0; x < 25; x++)
			{
				while (!(UCSR0A & _BV(UDRE0))) ;
				asm volatile(
						"ld r30, %a1"				"\n\t"
						"ldi r31, hi8(chartable)"	"\n\t"
						"add r31, %2"				"\n\t"
						"lpm %0, Z"
						: "=r" (out)
						: "e" (text), "r" (char_line)
						: "r30", "r31"
						);
				UDR0 = ~(out);
				text++;
			}
			break;
		case S_GRAPH:
			// draw graph data, or white line if at trigger
			if (video_line-V_GRAPH == y_trigger)
			{
				for (x = 0; x < 25; x++)
				{
					while (!(UCSR0A & _BV(UDRE0))) ;
					UDR0 = 0;
				}

			}
			else
			{
				for (x = 0; x < 25; x++)
				{
					while (!(UCSR0A & _BV(UDRE0))) ;
					UDR0 = ~(graph[graph_offset++]);
				}
			}
			break;
		case S_RULE:
			// draw a white line across the entire screen
			for (x = 0; x < 25; x++)
			{
				while (!(UCSR0A & _BV(UDRE0))) ;
				UDR0 = 0x00;
			}
			break;
		case S_BLANK:
			// do nothing
			break;
		}
	}

	// start new frame after line 262
	if (++video_line==263)
		video_line = 1;

	// set the screen section for next line
	if (video_line >= V_TOP && video_line < V_BOTTOM)
	{
		if (video_line < V_HEADER_SP)
		{
			section = S_TEXT;
			text = header;
			char_line = video_line-V_HEADER;
		}
		else if (video_line < V_HEADER_R)
			section = S_BLANK;
		else if (video_line < V_GRAPH)
			section = S_RULE;
		else if (video_line < V_FOOTER_R)
		{
			section = S_GRAPH;
			graph_offset = (video_line-V_GRAPH)*25;
		}
		else if (video_line < V_FOOTER_SP)
			section = S_RULE;
		else if (video_line < V_FOOTER)
			section = S_BLANK;
		else
		{
			section = S_TEXT;
			text = footer;
			char_line = video_line-V_FOOTER;
		}
	}
	else
	{
		section = S_BLANK;
	}
}

// function prototypes
void verify();
void initialize();
void drawPoint(uint8_t x, uint8_t y, uint8_t c);
void drawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t c);
void drawCursor(uint8_t x, uint8_t c);

int main(void)
{
	uint8_t x = 0;
	uint8_t in;
	uint8_t mode = 0;
	uint8_t char_count = 0;

	verify();
	initialize();

	// turn on interrupts
	sei();

	while (1)
	{
		// wait for REQ to be pulled low, set ACC low
		while (PINB & _BV(PINB3)) ;
		PORTB &= ~(_BV(PORTB2));

		// wait for reception to be complete, store byte
		while (!(SPSR & _BV(SPIF))) ;
		in = SPDR;

		// wait for REQ to be pulled high, set ACC high
		while (!(PINB & _BV(PINB3))) ;
		PORTB |= _BV(PORTB2);

		// bling LED
		PIND = _BV(PIND7);

		if (in & 0x80 && mode != 3 && mode != 4)
		{
			// command received

			switch (in & 0xF0)
			{
			case 0x80:
				// reset x-counter
				x = 0;
				break;
			case 0x90:
				// clear screen
				for (x = 0; x < 200; x++)
				{
					if (x < 199)
						drawLine(x, samples[x], x+1, samples[x+1], 0);

					samples[x] = 0;
				}

				// redraw cursor, if its visible
				if (x_cursor != 255)
					drawCursor(x_cursor, 1);

				// reset x-counter
				x = 0;
				break;
			case 0xA0:
				// change mode
				if ((in & 0x0F) == 0)
					mode = 0;
				else if ((in & 0x0F) == 1)
				{
					mode = 1;
					char_count = 0;
				}
				else if ((in & 0x0F) == 2)
				{
					mode = 2;
					char_count = 0;
				}
				else if ((in & 0x0F) == 3)
					mode = 3;
				else if ((in & 0x0F) == 4)
					mode = 4;
				break;
			default:
				break;
			}
		}
		else if (mode == 1)
		{
			// store character to header
			if (char_count < 25)
				header[char_count++] = in;
		}
		else if (mode == 2)
		{
			// store character to footer
			if (char_count < 25)
				footer[char_count++] = in;
		}
		else if (mode == 3)
		{
			// erase previous cursor, if it was visible
			if (x_cursor != 255)
			{
				drawCursor(x_cursor, 0);

				// fix the waveform
				if (x_cursor > 0)
				{
					drawLine(x_cursor-1, samples[x_cursor-1],
							x_cursor, samples[x_cursor], 1);
				}
				if (x_cursor < 199)
				{
					drawLine(x_cursor, samples[x_cursor], x_cursor+1,
							samples[x_cursor+1], 1);
				}
			}

			// store new cursor
			x_cursor = in;

			// draw new cursor, if it is visible
			if (x_cursor != 255)
				drawCursor(x_cursor, 1);

			// Always return to mode 0!
			mode = 0;
		}
		else if (mode == 4)
		{
			// store new trigger
			y_trigger = in;

			// Always return to mode 0!
			mode = 0;
		}
		else
		{
			if (x < 200)
			{
				// erase old line
				if (x < 199)
					drawLine(x, samples[x], x+1, samples[x+1], 0);

				// store new sample
				samples[x] = in & 0x7F;

				// draw new line
				if (x > 0)
					drawLine(x-1, samples[x-1], x, samples[x], 1);

				// redraw cursor
				if (x_cursor != 255)
					drawCursor(x_cursor, 1);

				// increment x-counter
				x++;
			}
		}
	}

	return 0;
}

/*
 * DRAWPOINT: Draws a point at (x,y) color c: 0=black, 1=white, 2=invert
 */
void drawPoint(uint8_t x, uint8_t y, uint8_t c)
{
	asm volatile(
			// i = (x>>3) + (y*25)
			// the byte with the pixel in it

			// get y
			"mov r16, %1"			"\n\t"

			// mult y by 25
			"ldi r17, 25"			"\n\t"
			"mul r16, r17"			"\n\t"

			// get x
			"mov r16, %0"			"\n\t"

			// divide x by 8
			"lsr r16"				"\n\t"
			"lsr r16"				"\n\t"
			"lsr r16"				"\n\t"

			// add in x/8
			"clr r17"				"\n\t"
			"add r16, r0"			"\n\t"
			"adc r17, r1"			"\n\t"

			"clr __zero_reg__"		"\n\t"

			// move screen to X (r27:r26)
			"mov r26, %A3"			"\n\t"
			"mov r27, %B3"			"\n\t"
			"add r26, r16"			"\n\t"
			"adc r27, r17"			"\n\t"

			// get screen byte
			"ld r5, X"				"\n\t"

			// form x & 7
			"mov r30, %0"			"\n\t"
			"andi r30, 0x07"		"\n\t"

			// get bit mask
			"clr r31"				"\n\t"
			"add r30, %A4"			"\n\t"
			"adc r31, %B4"			"\n\t"
			"lpm r6, Z"				"\n\t"

			// if (v4==1) graph[i] = v2 | v3 ;
			// if (v4==0) graph[i] = v2 & ~v3;
			// if (v4==2) graph[i] = v2 ^ v3 ;
			"cpi %2, 1"				"\n\t"
			"brne 1f"				"\n\t"
			"or  r5, r6"			"\n\t"
			"1:"					"\n\t"
			"cpi %2, 0"				"\n\t"
			"brne 2f"				"\n\t"
			"com r6"				"\n\t"
			"and r5, r6"			"\n\t"
			"2:"					"\n\t"
			"cpi %2, 2"				"\n\t"
			"brne 3f"				"\n\t"
			"eor r5, r6"			"\n\t"
			"3:"					"\n\t"

			// write the byte back to the screen
			"st X, r5"
			:
			: "r" (x), "r" (y), "d" (c), "r" (graph), "r" (bitmask)
			: "r5", "r6", "r16", "r17", "r26", "r27", "r30", "r31"
	);
}

/*
 * DRAWLINE: draw a line from (x1,y1) to (x2,y2), color c: 0=black, 1=white,
 * 		2=invert
 *
 * Code is from David Rodgers,
 * "Procedural Elements of Computer Graphics",1985
 */
void drawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t c)
{
	int16_t e;
	uint8_t dx,dy,j, temp;
	int8_t s1,s2;
	uint8_t x,y, xchange;

	x = x1;
	y = y1;
	dx = (x1 < x2) ? x2-x1 : x1-x2;
	dy = (y1 < y2) ? y2-y1 : y1-y2;
	s1 = (x1 == x2) ? 0 : ((x1 < x2) ? 1 : -1);
	s2 = (y1 == y2) ? 0 : ((y1 < y2) ? 1 : -1);
	xchange = 0;
	if (dy>dx)
	{
		temp = dx;
		dx = dy;
		dy = temp;
		xchange = 1;
	}
	e = ((int16_t)dy<<1) - dx;
	for (j=0; j<=dx; j++)
	{
		drawPoint(x,y,c) ;
		if (e>=0)
		{
			if (xchange==1) x = x + s1;
			else y = y + s2;
			e = e - ((int16_t)dx<<1);
		}
		if (xchange==1) y = y + s2;
		else x = x + s1;
		e = e + ((int16_t)dy<<1);
	}
}

/*
 * DRAWCURSOR: Draws the cursor at position x, color c: 0=black, 1=white,
 * 		2=invert
 */
void drawCursor(uint8_t x, uint8_t c)
{
	uint8_t y;

	for (y = 0; y < 128; y++)
		drawPoint(x, y, c);
}

/*
 * INITIALIZE: Set up the microcontroller registers and initialize global
 * 		variables.
 */
void initialize()
{
	uint8_t i;

	// PB7: SCK			in
	// PB6: MISO		out
	// PB5: MOSI		in
	// PB4: VIDEO CS	in
	// PB3: VIDEO REQ	in
	// PB2: VIDEO ACC	out
	// PB1: (unused)
	// PB0: USART XCK	out
	// Note: the USART XCK needs to be configured as an output even though it
	// 		is unconnected (required by MSPIM mode)
	PORTB = _BV(PORTB2);
	DDRB = _BV(DDB6) | _BV(DDB2) | _BV(DDB0);

	// PD7: LED			out
	// PD6: (unused)
	// PD5: (unused)
	// PD4: uC DETECT	in
	// PD3: (unused)
	// PD2: SYNC		out
	// PD1: VIDEO		out
	// PD0: USART RX	in
	DDRD = _BV(DDD7) | _BV(DDD2) | _BV(DDD1);

	// SPI: MSB first, slave, mode 00, fosc/4 (5 MHz)
	SPCR = _BV(SPE);

	// TIMER 1: OC1* disconnected, CTC mode, fosc/1 (20MHz), OC1A and OC1B
	//		interrupts enabled
	TCCR1B = _BV(WGM12) | _BV(CS10);
	OCR1A = LINE_TIME;	// time for one NTSC line
	OCR1B = SLEEP_TIME;	// time to go to sleep
	TIMSK1 = _BV(OCIE1B) | _BV(OCIE1A);

	// USART in MSPIM mode, transmitter enabled, frequency fosc/4
	UCSR0B = _BV(TXEN0);
	UCSR0C = _BV(UMSEL01) | _BV(UMSEL00);
	UBRR0 = 1;

	// SLEEP: idle sleep mode
	SMCR = 0x01;

	// cursor and trigger off
	x_cursor = 255;
	y_trigger = 255;

	// initialize the sample array
	for (i = 0; i < 200; i++)
		samples[i] = 0;

	// start at first videoline, with regular sync pulses
	video_line = 1;
	sync_on = 0;
	sync_off = _BV(PORTD2);
}

/*
 * VERIFY: Check to make sure the code is running on the correct
 * 		microcontroller. If not, flash a warning on the LED forever.
 */
void verify()
{
	uint8_t i;

	// if pin 4 on port D is low, everything's fine!
	if (!(PIND & _BV(PIND4)))
		return;

	// we're executing on the wrong microcontroller!
	DDRD |= _BV(DDD7);
	while (1)
	{
		// flash the LED forever
		PORTD |= _BV(PD7);
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD &= ~(_BV(PD7));
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD |= _BV(PD7);
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD &= ~(_BV(PD7));
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD |= _BV(PD7);
		for (i = 0; i < 20; i++)
			_delay_ms(10);
		PORTD &= ~(_BV(PD7));
		for (i = 0; i < 100; i++)
			_delay_ms(10);
	}
}
